from __future__ import annotations
import asyncio
import contextlib
import functools
import gc
import inspect
import io
import os
import pickle
import sys
import textwrap
import types
from tempfile import TemporaryDirectory
import pytest
from line_profiler import _line_profiler, LineProfiler, LineStats


def f(x):
    """A docstring."""
    y = x + 10
    return y


def g(x):
    y = yield x + 10
    yield y + 20


async def ag(delay, start=()):
    i = 0
    for x in start:
        yield i
        i += x
    while True:
        received = await asyncio.sleep(delay, (yield i))
        if received is None:
            return
        i += received


def get_profiling_tool_name():
    return sys.monitoring.get_tool(sys.monitoring.PROFILER_ID)


def strip(s):
    return textwrap.dedent(s).strip('\n')


def get_prof_stats(prof, name='prof', **kwargs):
    with io.StringIO() as sio:
        prof.print_stats(sio, **kwargs)
        output = sio.getvalue()
        print(f'@{name}:', textwrap.indent(output, '  '), sep='\n\n')
    return output


class check_timings_and_mem:
    """
    Verify that:
    - The profiler starts without timing data and ends with some.
    - We don't leak reference counts for code objects (which are
      retrieved with a C function in the Cython code).
    """

    def __init__(
        self, prof, *, check_timings=True, check_ref_counts=True, gc=False
    ):
        self.prof = prof
        self.check_timings = bool(check_timings)
        self.check_ref_counts = bool(check_ref_counts)
        self.ref_counts = None
        # Sometimes the garbage collector seem to need extra help;
        # in those cases, explicitly garbage-collect
        self.gc = gc

    def __enter__(self):
        self._check_timings_enter()
        self._check_ref_counts_enter()
        return self.prof

    def __exit__(self, *_, **__):
        self._check_timings_exit()
        self._check_ref_counts_exit()

    def _check_timings_enter(self):
        if not self.check_timings:
            return
        timings = self.timings
        assert not any(timings.values()), (
            f'Expected no timing entries, got {timings!r}'
        )

    def _check_timings_exit(self):
        if not self.check_timings:
            return
        timings = self.timings
        assert any(timings.values()), (
            f'Expected timing entries, got {timings!r}'
        )

    def _check_ref_counts_enter(self):
        if not self.check_ref_counts:
            return
        self.ref_counts = self.get_ref_counts()

    def _check_ref_counts_exit(self):
        if not self.check_ref_counts:
            return
        assert self.ref_counts is not None
        for key, count in self.get_ref_counts().items():
            try:
                referrers = repr(
                    gc.get_referrers(
                        *(
                            code
                            for code in self.prof.code_hash_map
                            if code.co_name == key
                        )
                    )
                )
                msg = (
                    f'{key}(): '
                    f'ref count {self.ref_counts[key]} -> {count} '
                    f'(referrers: {referrers})'
                )
                if self.ref_counts[key] == count:
                    print(msg)
                else:
                    raise AssertionError(msg)
            except KeyError:
                pass

    def get_ref_counts(self):
        if self.gc:
            gc.collect()
        results = {}
        for code in self.prof.code_hash_map:
            # Note: use the name as the key to avoid messing with the
            # ref count
            key = code.co_name
            results[key] = results.get(key, 0) + sys.getrefcount(code)
        return results

    @property
    def timings(self):
        return self.prof.get_stats().timings


def test_init():
    lp = LineProfiler()
    assert lp.functions == []
    assert lp.code_map == {}
    lp = LineProfiler(f)
    assert lp.functions == [f]
    assert lp.code_map == {f.__code__: {}}
    lp = LineProfiler(f, g)
    assert lp.functions == [f, g]
    assert lp.code_map == {
        f.__code__: {},
        g.__code__: {},
    }


def test_last_time():
    """
    Test that `LineProfiler.c_last_time` and `LineProfiler.last_time`
    are consistent.
    """
    prof = LineProfiler()
    with pytest.raises(KeyError, match='[Nn]o profiling data'):
        prof.c_last_time

    def get_last_time(prof, *, c=False):
        try:
            return getattr(prof, 'c_last_time' if c else 'last_time')
        except KeyError:
            return {}

    @prof
    def func():
        return (get_last_time(prof, c=True).copy(), get_last_time(prof).copy())

    # These are always empty outside a profiling context
    # (hence the need of the above function to capture the transient
    # values)
    assert not get_last_time(prof, c=True)
    assert not get_last_time(prof)
    # Inside `func()`, both should get an entry therefor
    clt, lt = func()
    assert not get_last_time(prof, c=True)
    assert not get_last_time(prof)
    assert set(clt) == {hash(func.__wrapped__.__code__.co_code)}
    assert set(lt) == {func.__wrapped__.__code__}


def test_enable_disable():
    lp = LineProfiler()
    assert lp.enable_count == 0
    lp.enable_by_count()
    assert lp.enable_count == 1
    lp.enable_by_count()
    assert lp.enable_count == 2
    lp.disable_by_count()
    assert lp.enable_count == 1
    lp.disable_by_count()
    assert lp.enable_count == 0
    assert lp.last_time == {}
    lp.disable_by_count()
    assert lp.enable_count == 0

    with lp:
        assert lp.enable_count == 1
        with lp:
            assert lp.enable_count == 2
        assert lp.enable_count == 1
    assert lp.enable_count == 0
    assert lp.last_time == {}

    with pytest.raises(RuntimeError):
        assert lp.enable_count == 0
        with lp:
            assert lp.enable_count == 1
            raise RuntimeError()
    assert lp.enable_count == 0
    assert lp.last_time == {}


def test_double_decoration():
    """
    Test that wrapping the same function twice does not result in
    spurious profiling entries.
    """
    profile = LineProfiler()
    f_wrapped = profile(f)
    f_double_wrapped = profile(f_wrapped)
    assert f_double_wrapped is f_wrapped

    with check_timings_and_mem(profile):
        assert profile.enable_count == 0
        value = f_wrapped(10)
        assert profile.enable_count == 0
        assert value == f(10)

    assert len(profile.get_stats().timings) == 1


def test_function_decorator():
    """
    Test for `LineProfiler.wrap_function()`.
    """
    profile = LineProfiler()
    f_wrapped = profile(f)
    assert f in profile.functions
    assert f_wrapped.__name__ == 'f'

    with check_timings_and_mem(profile):
        assert profile.enable_count == 0
        value = f_wrapped(10)
        assert profile.enable_count == 0
        assert value == f(10)


def test_gen_decorator():
    """
    Test for `LineProfiler.wrap_generator()`.
    """
    profile = LineProfiler()
    g_wrapped = profile(g)
    assert inspect.isgeneratorfunction(g_wrapped)
    assert g in profile.functions
    assert g_wrapped.__name__ == 'g'

    with check_timings_and_mem(profile):
        assert profile.enable_count == 0
        i = g_wrapped(10)
        assert profile.enable_count == 0
        assert next(i) == 20
        assert profile.enable_count == 0
        assert i.send(30) == 50
        assert profile.enable_count == 0
        with pytest.raises(StopIteration):
            next(i)
        assert profile.enable_count == 0


def test_coroutine_decorator():
    """
    Test for `LineProfiler.wrap_coroutine()`.
    """

    async def coro(delay=0.015625):
        return await asyncio.sleep(delay, 1)

    profile = LineProfiler()
    coro_wrapped = profile(coro)
    assert inspect.iscoroutinefunction(coro)
    assert coro in profile.functions

    with check_timings_and_mem(profile):
        assert profile.enable_count == 0
        assert asyncio.run(coro_wrapped()) == 1
        assert profile.enable_count == 0


@pytest.mark.parametrize('gc', [True, False])
def test_async_gen_decorator(gc):
    """
    Test for `LineProfiler.wrap_async_generator()`.
    """
    delay = 0.015625

    async def use_agen_complex(*args, delay=delay):
        results = []
        agen = ag_wrapped(delay)
        results.append(await agen.asend(None))  # Start the generator
        for send in args:
            with (
                pytest.raises(StopAsyncIteration)
                if send is None
                else contextlib.nullcontext()
            ):
                results.append(await agen.asend(send))
            if send is None:
                break
        return results

    async def use_agen_simple(*args, delay=delay):
        results = []
        async for i in ag_wrapped(delay, args):
            results.append(i)
        return results

    def has_refcount_bug():
        """
        Apparently 3.12 introduced a refcount bug involving async
        generators (CPython issue GH-100964), which is *finally*
        patched in CPython GitHub PR #141112.
        As such, we have to check the patch versions to determine
        whether refcount inconsistencies are to be expected;
        alternatively, inspect `test.test_generators.FinalizationTest`
        for the test method `.test_exhausted_generator_frame_cycle()`;
        but that seems even more fragile than checking the versions...
        """
        version = sys.version_info[:2]
        patch = sys.version_info[2]
        if version == (3, 12):  # 3.12.x all bugged
            return True
        if version == (3, 13):  # Fix backported to 3.13.12...
            return patch < 12
        if version == (3, 14):  # ... and 3.14.3
            return patch < 3
        return False

    profile = LineProfiler()
    ag_wrapped = profile(ag)
    xfail_refcount = has_refcount_bug() and not gc
    assert inspect.isasyncgenfunction(ag_wrapped)
    assert ag in profile.functions

    with check_timings_and_mem(profile):
        assert profile.enable_count == 0
        assert asyncio.run(use_agen_simple()) == [0]
        assert profile.enable_count == 0
        assert asyncio.run(use_agen_simple(1, 2, 3)) == [0, 1, 3, 6]
        assert profile.enable_count == 0
    with contextlib.ExitStack() as stack:
        if xfail_refcount:
            excinfo = stack.enter_context(
                pytest.raises(AssertionError, match=r'ag\(\): ref count')
            )
        stack.enter_context(
            check_timings_and_mem(profile, check_timings=False, gc=gc)
        )
        assert profile.enable_count == 0
        assert asyncio.run(use_agen_complex(1, 2, 3)) == [0, 1, 3, 6]
        assert profile.enable_count == 0
        assert asyncio.run(use_agen_complex(1, 2, 3, None, 4)) == [0, 1, 3, 6]
        assert profile.enable_count == 0
    if xfail_refcount:
        pytest.xfail(
            '\nsys.version={!r}..., gc={}:\n{}'.format(
                sys.version.strip().split()[0], gc, excinfo.getrepr(style='no')
            )
        )


def test_classmethod_decorator():
    """
    Test for `LineProfiler.wrap_classmethod()`.

    Notes
    -----
    This is testing for an edge case;
    for the best result, always use `@profile` as the innermost
    decorator, as auto-profile normally does.
    """
    profile = LineProfiler()

    class Object:
        @profile
        @classmethod
        def foo(cls) -> str:
            return cls.__name__ * 2

    assert isinstance(inspect.getattr_static(Object, 'foo'), classmethod)
    assert profile.enable_count == 0
    assert len(profile.functions) == 1
    assert Object.foo() == Object().foo() == 'ObjectObject'
    output = strip(get_prof_stats(profile, name='profile', summarize=True))
    # Check that we have profiled `Object.foo()`
    assert output.endswith('foo')
    (line,) = (line for line in output.splitlines() if line.endswith('* 2'))
    # Check that it has been run twice
    assert int(line.split()[1]) == 2
    assert profile.enable_count == 0


def test_staticmethod_decorator():
    """
    Test for `LineProfiler.wrap_staticmethod()`.

    Notes
    -----
    This is testing for an edge case;
    for the best result, always use `@profile` as the innermost
    decorator, as auto-profile normally does.
    """
    profile = LineProfiler()

    class Object:
        @profile
        @staticmethod
        def foo(x: int) -> int:
            return x * 2

    assert isinstance(inspect.getattr_static(Object, 'foo'), staticmethod)
    assert profile.enable_count == 0
    assert len(profile.functions) == 1
    assert Object.foo(3) == Object().foo(3) == 6
    output = strip(get_prof_stats(profile, name='profile', summarize=True))
    # Check that we have profiled `Object.foo()`
    assert output.endswith('foo')
    (line,) = (line for line in output.splitlines() if line.endswith('* 2'))
    # Check that it has been run twice
    assert int(line.split()[1]) == 2
    assert profile.enable_count == 0


def test_boundmethod_decorator():
    """
    Test for `LineProfiler.wrap_boundmethod()`.

    Notes
    -----
    This is testing for an edge case;
    for the best result, always use `@profile` as the innermost
    decorator, as auto-profile normally does.
    """
    profile = LineProfiler()

    class Object:
        def foo(self, x: int) -> int:
            return id(self) * x

    obj = Object()
    # Check that calls are aggregated
    profiled_foo_1 = profile(obj.foo)
    profiled_foo_2 = profile(obj.foo)
    assert isinstance(profiled_foo_1, types.MethodType)
    assert isinstance(profiled_foo_2, types.MethodType)
    assert profile.enable_count == 0
    # XXX: should we try do remove duplicates?
    assert profile.functions == [Object.foo, Object.foo]
    assert profiled_foo_1(2) == profiled_foo_2(2) == obj.foo(2) == id(obj) * 2
    output = strip(get_prof_stats(profile, name='profile', summarize=True))
    # Check that we have profiled `Object.foo()`
    assert output.endswith('foo')
    (line,) = (line for line in output.splitlines() if line.endswith('* x'))
    # Check that the wrapped methods has been run twice in total
    assert int(line.split()[1]) == 2
    assert profile.enable_count == 0


def test_partialmethod_decorator():
    """
    Test for `LineProfiler.wrap_partialmethod()`

    Notes
    -----
    This is testing for an edge case;
    for the best result, always use `@profile` as the innermost
    decorator in a function definition, as auto-profile normally does.
    """
    profile = LineProfiler()

    class Object:
        def foo(self, x: int) -> int:
            return id(self) * x

        bar = profile(functools.partialmethod(foo, 1))

    assert isinstance(
        inspect.getattr_static(Object, 'bar'), functools.partialmethod
    )
    obj = Object()
    assert profile.enable_count == 0
    assert profile.functions == [Object.foo]
    assert obj.foo(1) == obj.bar() == id(obj)
    output = strip(get_prof_stats(profile, name='profile', summarize=True))
    # Check that we have profiled `Object.foo()` (via `.bar()`)
    assert output.endswith('foo')
    (line,) = (line for line in output.splitlines() if line.endswith('* x'))
    # Check that the wrapped method has been run once
    assert int(line.split()[1]) == 1
    assert profile.enable_count == 0


def test_partial_decorator() -> None:
    """
    Test for `LineProfiler.wrap_partial()`.

    Notes
    -----
    This is testing for an edge case;
    for the best result, always use `@profile` as the innermost
    decorator, as auto-profile normally does.
    """
    profile = LineProfiler()

    def foo(x: int, y: int) -> int:
        return x + y

    bar = functools.partial(foo, 2)
    profiled_bar_1 = profile(bar)
    profiled_bar_2 = profile(bar)
    assert isinstance(profiled_bar_1, functools.partial)
    assert isinstance(profiled_bar_2, functools.partial)
    assert profile.enable_count == 0
    # XXX: should we try do remove duplicates?
    assert profile.functions == [foo, foo]
    assert profiled_bar_1(3) == profiled_bar_2(3) == bar(3) == foo(2, 3) == 5
    output = strip(get_prof_stats(profile, name='profile', summarize=True))
    # Check that we have profiled `foo()`
    assert output.endswith('foo')
    (line,) = (line for line in output.splitlines() if line.endswith('x + y'))
    # Check that the wrapped partials has been run twice in total
    assert int(line.split()[1]) == 2
    assert profile.enable_count == 0


def test_property_decorator():
    """
    Test for `LineProfiler.wrap_property()`.

    Notes
    -----
    This is testing for an edge case;
    for the best result, always use `@profile` as the innermost
    decorator, as auto-profile normally does.
    """
    profile = LineProfiler()

    class Object:
        def __init__(self, x: int) -> None:
            self.x = x

        @profile
        @property
        def foo(self) -> int:
            return self.x * 2

        # The profiler sees both the setter and the already-wrapped
        # getter here, but it shouldn't re-wrap the getter

        @profile
        @foo.setter
        def foo(self, foo) -> None:
            self.x = foo // 2

    assert isinstance(Object.foo, property)
    assert profile.enable_count == 0
    assert len(profile.functions) == 2
    obj = Object(3)
    assert obj.foo == 6  # Use getter
    obj.foo = 10  # Use setter
    assert obj.x == 5
    assert obj.foo == 10  # Use getter
    output = strip(get_prof_stats(profile, name='profile', summarize=True))
    # Check that we have profiled `Object.foo`
    assert output.endswith('foo')
    (getter_line,) = (
        line for line in output.splitlines() if line.endswith('* 2')
    )
    (setter_line,) = (
        line for line in output.splitlines() if line.endswith('// 2')
    )
    # Check that the getter has been run twice and the setter once
    assert int(getter_line.split()[1]) == 2
    assert int(setter_line.split()[1]) == 1
    assert profile.enable_count == 0


def test_cached_property_decorator():
    """
    Test for `LineProfiler.wrap_cached_property()`.

    Notes
    -----
    This is testing for an edge case;
    for the best result, always use `@profile` as the innermost
    decorator, as auto-profile normally does.
    """
    profile = LineProfiler()

    class Object:
        def __init__(self, x: int) -> None:
            self.x = x

        @profile
        @functools.cached_property
        def foo(self) -> int:
            return self.x * 2

    assert isinstance(Object.foo, functools.cached_property)
    assert profile.enable_count == 0
    assert len(profile.functions) == 1
    obj = Object(3)
    assert obj.foo == 6  # Use getter
    assert obj.foo == 6  # Getter not called because it's cached
    output = strip(get_prof_stats(profile, name='profile', summarize=True))
    # Check that we have profiled `Object.foo`
    assert output.endswith('foo')
    (line,) = (line for line in output.splitlines() if line.endswith('* 2'))
    # Check that the getter has been run once
    assert int(line.split()[1]) == 1
    assert profile.enable_count == 0


def test_class_decorator():
    """
    Test for `LineProfiler.wrap_class()`.
    """
    profile = LineProfiler()

    def unrelated(x):
        return str(x)

    @profile
    class Object:
        def __init__(self, x):
            self.x = self.convert(x)

        @property
        def id(self):
            return id(self)

        @classmethod
        def class_method(cls, n):
            return cls.__name__ * n

        # This is unrelated to `Object` and shouldn't be profiled
        convert = staticmethod(unrelated)

    # Are we keeping tabs on the correct entities?
    assert len(profile.functions) == 3
    assert set(profile.functions) == {
        Object.__init__.__wrapped__,
        Object.id.fget.__wrapped__,
        vars(Object)['class_method'].__func__.__wrapped__,
    }
    # Make some calls
    assert not profile.enable_count
    obj = Object(1)
    assert obj.x == '1'
    assert id(obj) == obj.id
    assert obj.class_method(3) == 'ObjectObjectObject'
    assert not profile.enable_count
    # Check the profiling results
    all_entries = sorted(sum(profile.get_stats().timings.values(), []))
    assert len(all_entries) == 3
    assert all(nhits == 1 for (_, nhits, _) in all_entries)


def test_add_class_wrapper():
    """
    Test adding a callable-wrapper object wrapping a class.
    """
    profile = LineProfiler()

    class Object:
        @classmethod
        class method:
            def __init__(self, cls, x):
                self.cls = cls
                self.x = x

            def __repr__(self):
                fmt = '{.__name__}.{.__name__}({!r})'.format
                return fmt(self.cls, type(self), self.x)

    # Bookkeeping
    profile.add_class(Object)
    method_cls = vars(Object)['method'].__func__
    assert profile.functions == [method_cls.__init__, method_cls.__repr__]
    # Actual profiling
    with profile:
        obj = Object.method(1)
        assert obj.cls == Object
        assert obj.x == 1
        assert repr(obj) == 'Object.method(1)'
    # Check data
    all_nhits = {
        func_name.rpartition('.')[-1]: sum(nhits for (_, nhits, _) in entries)
        for (*_, func_name), entries in profile.get_stats().timings.items()
    }
    assert all_nhits['__init__'] == all_nhits['__repr__'] == 2


@pytest.mark.parametrize('decorate', [True, False])
def test_profiler_c_callable_no_op(decorate):
    """
    Test that the following are no-ops on C-level callables:
    - Decoration (`.__call__()`): the callable is returned as-is.
    - `.add_callable()`: it returns 0.
    """
    profile = LineProfiler()

    for func, Type in [
        (len, types.BuiltinFunctionType),
        ('string'.split, types.BuiltinMethodType),
        (vars(int)['from_bytes'], types.ClassMethodDescriptorType),
        (str.split, types.MethodDescriptorType),
        ((1).__str__, types.MethodWrapperType),
        (int.__repr__, types.WrapperDescriptorType),
    ]:
        assert isinstance(func, Type)
        if decorate:  # Decoration is no-op
            assert profile(func) is func
        else:  # Add is no-op
            assert not profile.add_callable(func)
        assert not profile.functions


def test_show_func_column_formatting():
    from line_profiler.line_profiler import show_func
    import line_profiler
    import io

    # Use a function in this module as an example
    func = line_profiler.line_profiler.show_text
    start_lineno = func.__code__.co_firstlineno
    filename = func.__code__.co_filename
    func_name = func.__name__

    def get_func_linenos(func):
        import sys

        if sys.version_info[0:2] >= (3, 10):
            return sorted(
                set(
                    [
                        t[0] if t[2] is None else t[2]
                        for t in func.__code__.co_lines()
                    ]
                )
            )
        else:
            import dis

            return sorted(
                set([t[1] for t in dis.findlinestarts(func.__code__)])
            )

    line_numbers = get_func_linenos(func)

    unit = 1.0
    output_unit = 1.0
    stripzeros = False

    # Build fake timeings for each line in the example function
    timings = [
        (lineno, idx * 1e13, idx * (2e10 ** (idx % 3)))
        for idx, lineno in enumerate(line_numbers, start=1)
    ]
    stream = io.StringIO()
    show_func(
        filename,
        start_lineno,
        func_name,
        timings,
        unit,
        output_unit,
        stream,
        stripzeros,
    )
    text = stream.getvalue()
    print(text)

    timings = [
        (lineno, idx * 1e15, idx * 2e19)
        for idx, lineno in enumerate(line_numbers, start=1)
    ]
    stream = io.StringIO()
    show_func(
        filename,
        start_lineno,
        func_name,
        timings,
        unit,
        output_unit,
        stream,
        stripzeros,
    )
    text = stream.getvalue()
    print(text)

    # TODO: write a check to verify columns are aligned nicely


@pytest.mark.skipif(
    not hasattr(sys, 'monitoring'),
    reason='no `sys.monitoring` in version '
    f'{".".join(str(v) for v in sys.version_info[:2])}',
)
def test_sys_monitoring(monkeypatch):
    """
    Test that `LineProfiler` is properly registered with
    `sys.monitoring`.
    """
    monkeypatch.setattr(_line_profiler, 'USE_LEGACY_TRACE', False)
    profile = LineProfiler()
    get_name_wrapped = profile(get_profiling_tool_name)
    tool = get_profiling_tool_name()
    assert tool is None, (
        f'Expected no active profiling tool before profiling, got {tool!r}'
    )
    tool = get_name_wrapped()
    assert tool == 'line_profiler', (
        "Expected 'line_profiler' to be registered with `sys.monitoring` "
        f'when a profiled function is run, got {tool!r}'
    )
    tool = get_profiling_tool_name()
    assert tool is None, (
        f'Expected no active profiling tool after profiling, got {tool!r}'
    )


def test_profile_generated_code():
    """
    Test that profiling shows source information with generated code.
    """
    import linecache
    from line_profiler import LineProfiler
    from line_profiler.line_profiler import is_generated_code

    # Simulate generated code in linecache

    # Note: this test will fail if the generated code name does not
    # start with "<generated: ".
    generated_code_name = "<generated: 'test_fn'>"
    assert is_generated_code(generated_code_name)

    code_lines = ['def test_fn():', '    return 42']

    linecache.cache[generated_code_name] = (
        None,
        None,
        [l + '\n' for l in code_lines],
        None,
    )

    # Compile the generated code
    ns = {}
    exec(
        compile(
            ''.join(l + '\n' for l in code_lines), generated_code_name, 'exec'
        ),
        ns,
    )
    fn = ns['test_fn']

    # Profile the generated function
    profiler = LineProfiler()
    profiled_fn = profiler(fn)
    profiled_fn()

    output = get_prof_stats(profiler, 'profiler')

    # Check that the output contains the generated code's source lines
    for line in code_lines:
        assert line in output

    # .. as well as the generated code name
    assert generated_code_name in output


def test_multiple_profilers_usage():
    """
    Test using more than one profilers simultaneously.
    """
    prof1 = LineProfiler()
    prof2 = LineProfiler()

    def sum_n(n):
        x = 0
        for n in range(1, n + 1):
            x += n
        return x

    @prof1
    def sum_n_sq(n):
        x = 0
        for n in range(1, n + 1):
            x += n**2
        return x

    @prof2
    def sum_n_cb(n):
        x = 0
        for n in range(1, n + 1):
            x += n**3
        return x

    # If we decorate a wrapper, just "register" the profiler with the
    # existing wrapper and add the wrapped function
    sum_n_wrapper_1 = prof1(sum_n)
    assert prof1.functions == [sum_n_sq.__wrapped__, sum_n]
    sum_n_wrapper_2 = prof2(sum_n_wrapper_1)
    assert sum_n_wrapper_2 is not sum_n_wrapper_1
    assert prof2.functions == [sum_n_cb.__wrapped__, sum_n]

    # Call the functions
    n = 400
    assert sum_n_wrapper_1(n) == 0.5 * n * (n + 1)
    assert sum_n_wrapper_2(n) == 0.5 * n * (n + 1)
    assert 6 * sum_n_sq(n) == n * (n + 1) * (2 * n + 1)
    assert sum_n_cb(n) == 0.25 * (n * (n + 1)) ** 2

    # Inspect the timings
    t1 = {
        fname.rpartition('.')[-1]: entries
        for (*_, fname), entries in prof1.get_stats().timings.items()
    }
    t2 = {
        fname.rpartition('.')[-1]: entries
        for (*_, fname), entries in prof2.get_stats().timings.items()
    }
    assert set(t1) == {'sum_n_sq', 'sum_n'}
    assert set(t2) == {'sum_n_cb', 'sum_n'}
    # Note: `prof1` active when both wrapper is called, but `prof2` only
    # when `sum_n_wrapper_2()` is
    assert t1['sum_n'][2][1] == 2 * n
    assert t2['sum_n'][2][1] == n
    assert t1['sum_n_sq'][2][1] == n
    assert t2['sum_n_cb'][2][1] == n


def test_duplicate_code_objects():
    """
    Test that results are correctly aggregated between duplicate code
    objects.
    """
    code = textwrap.dedent("""
    @profile
    def func(n):
        x = 0
        for n in range(1, n + 1):
            x += n
        return x
    """).strip('\n')
    profile = LineProfiler()
    # Create and call the function once
    namespace_1 = {'profile': profile}
    exec(code, namespace_1)
    assert 'func' in namespace_1
    assert len(profile.functions) == 1
    assert namespace_1['func'].__wrapped__ in profile.functions
    assert namespace_1['func'](10) == 10 * 11 // 2
    # Do it again
    namespace_2 = {'profile': profile}
    exec(code, namespace_2)
    assert 'func' in namespace_2
    assert len(profile.functions) == 2
    assert namespace_2['func'].__wrapped__ in profile.functions
    assert namespace_2['func'](20) == 20 * 21 // 2
    # Check that data from both calls are aggregated
    # (Entries are represented as tuples `(lineno, nhits, time)`)
    (entries,) = profile.get_stats().timings.values()
    assert entries[-2][1] == 10 + 20


@pytest.mark.parametrize('force_same_line_numbers', [True, False])
@pytest.mark.parametrize(
    'ops',
    [
        # Replication of the problematic case in issue #350
        'func1:prof_all'
        '-func2:prof_some:prof_all'
        '-func3:prof_all'
        '-func4:prof_some:prof_all',
        # Invert the order of decoration
        'func1:prof_all'
        '-func2:prof_all:prof_some'
        '-func3:prof_all'
        '-func4:prof_all:prof_some',
        # More profiler stacks
        'func1:p1:p2-func2:p2:p3-func3:p3:p4-func4:p4:p1',
        'func1:p1:p2:p3-func2:p2:p3:p4-func3:p3:p4:p1-func4:p4:p1:p2',
        'func1:p1:p2:p3-func2:p4:p3:p2-func3:p3:p4:p1-func4:p2:p1:p4',
        # Misc. edge cases
        # - Naive padding of the following case would cause `func1()`
        #   and `func2()` to end up with the same bytecode, so guard
        #   against it
        'func1:p1:p2'  # `func1()` padded once?
        '-func2:p3'  # `func2()` padded twice?
        '-func1:p4:p3',  # `func1()` padded once (again)?
        # - Check that double decoration doesn't mess things up
        'func1:p1:p2'
        '-func2:p2:p3'
        '-func3:p3:p4'
        '-func4:p4:p1'
        '-func1:p1',  # Now we're passing `func1()` to `p1` twice
    ],
)
def test_multiple_profilers_identical_bytecode(
    tmp_path, ops, force_same_line_numbers
):
    """
    Test that functions compiling down to the same bytecode are
    correctly handled between multiple profilers.

    Notes
    -----
    - `ops` should consist of chunks joined by hyphens, where each chunk
      has the format `<func_id>:<prof_name>[:<prof_name>[...]]`,
      indicating that the profilers are to be used in order to decorate
      the specified function.
    - `force_same_line_numbers` is used to coerce all functions to
      compile down to code objects with the same line numbers.
    """

    def check_seen(name, output, func_id, expected):
        lines = [
            line
            for line in output.splitlines()
            if line.startswith('Function: ')
        ]
        if any(func_id in line for line in lines) == expected:
            return
        if expected:
            raise AssertionError(f"profiler `@{name}` didn't see `{func_id}()`")
        raise AssertionError(f'profiler `@{name}` saw `{func_id}()`')

    def check_has_profiling_data(name, output, func_id, expected):
        assert func_id.startswith('func')
        nloops = func_id[len('func') :]
        try:
            line = next(
                line
                for line in output.splitlines()
                if line.endswith(f'result.append({nloops})')
            )
        except StopIteration:
            if expected:
                raise AssertionError(
                    f"profiler `@{name}` didn't see `{func_id}()`"
                )
            else:
                return
        if (line.split()[1] == nloops) == expected:
            return
        if expected:
            raise AssertionError(
                f"profiler `@{name}` didn't get data from `{func_id}()`"
            )
        raise AssertionError(f'profiler `@{name}` got data from `{func_id}()`')

    if force_same_line_numbers:
        funcs = {}
        pattern = strip("""
        def func{0}():
            result = []
            for _ in range({0}):
                result.append({0})
            return result
            """)
        for i in [1, 2, 3, 4]:
            tempfile = tmp_path / f'func{i}.py'
            source = pattern.format(i)
            tempfile.write_text(source)
            exec(compile(source, str(tempfile), 'exec'), funcs)
    else:

        def func1():
            result = []
            for _ in range(1):
                result.append(1)
            return result

        def func2():
            result = []
            for _ in range(2):
                result.append(2)
            return result

        def func3():
            result = []
            for _ in range(3):
                result.append(3)
            return result

        def func4():
            result = []
            for _ in range(4):
                result.append(4)
            return result

        funcs = {'func1': func1, 'func2': func2, 'func3': func3, 'func4': func4}

    # Apply the decorators in order
    all_dec_names = {f'func{i}': set() for i in [1, 2, 3, 4]}
    all_profs = {}
    for op in ops.split('-'):
        func_id, *profs = op.split(':')
        all_dec_names[func_id].update(profs)
        for name in profs:
            try:
                prof = all_profs[name]
            except KeyError:
                prof = all_profs[name] = LineProfiler()
            funcs[func_id] = prof(funcs[func_id])
    # Call each function once
    assert funcs['func1']() == [1]
    assert funcs['func2']() == [2, 2]
    assert funcs['func3']() == [3, 3, 3]
    assert funcs['func4']() == [4, 4, 4, 4]
    # Check that the bytecodes of the profiled functions are distinct
    profiled_funcs = {
        funcs[name].__line_profiler_id__.func
        for name, decs in all_dec_names.items()
        if decs
    }
    assert len({func.__code__.co_code for func in profiled_funcs}) == len(
        profiled_funcs
    )
    # Check the profiling results
    for name, prof in sorted(all_profs.items()):
        output = get_prof_stats(prof, name=name, summarize=True)
        for func_id, decs in all_dec_names.items():
            profiled = name in decs
            check_seen(name, output, func_id, profiled)
            check_has_profiling_data(name, output, func_id, profiled)


def test_aggregate_profiling_data_between_code_versions():
    """
    Test that profiling data from previous versions of the code object
    are preserved when another profiler causes the code object of a
    function to be overwritten.
    """

    def func(n):
        x = 0
        for n in range(1, n + 1):
            x += n
        return x

    prof1 = LineProfiler()
    prof2 = LineProfiler()

    # Gather data with `@prof1`
    wrapper1 = prof1(func)
    assert wrapper1(10) == 10 * 11 // 2
    code = func.__code__
    # Gather data with `@prof2`; the code object is overwritten here
    wrapper2 = prof2(wrapper1)
    assert func.__code__ != code
    assert wrapper2(15) == 15 * 16 // 2
    # Despite the overwrite of the code object, the old data should
    # still remain, and be aggregated with the new data when calling
    # `prof1.get_stats()`
    for prof, name, count in (prof1, 'prof1', 25), (prof2, 'prof2', 15):
        result = get_prof_stats(prof, name)
        loop_body = next(
            line for line in result.splitlines() if line.endswith('x += n')
        )
        assert loop_body.split()[1] == str(count)


@pytest.mark.xfail(
    condition=sys.version_info[:2] == (3, 9),
    reason='Handling of `finally` bugged in Python 3.9',
)
def test_profiling_exception():
    """
    Test that profiling data is reported for:
    - The line raising an exception
    - The last lines in the `except` and `finally` subblocks of a
      `try`-(`except`-)`finally` statement

    Notes
    -----
    Seems to be bugged for Python 3.9 only; may be related to CPython
    issue #83295.
    """
    prof = LineProfiler()

    class MyException(Exception):
        pass

    @prof
    def func_raise():
        pass
        raise MyException  # Raise: raise
        l.append(0)

    @prof
    def func_try_finally():
        try:
            raise MyException  # Try-finally: try
        finally:
            l.append(1)  # Try-finally: finally

    @prof
    def func_try_except_finally(reraise):
        try:
            raise MyException  # Try-except-finally: try
        except MyException:
            l.append(2)  # Try-except-finally: except
            if reraise:
                raise
        finally:
            l.append(3)  # Try-except-finally: finally

    l = []
    for func in [
        func_raise,
        func_try_finally,
        functools.partial(func_try_except_finally, True),
        functools.partial(func_try_except_finally, False),
    ]:
        try:
            func()
        except MyException:
            pass
    result = get_prof_stats(prof)
    assert l == [1, 2, 3, 2, 3]
    for stmt, nhits in [
        ('raise', 1),
        ('try-finally', 1),
        ('try-except-finally', 2),
    ]:
        for step in stmt.split('-'):
            comment = '# {}: {}'.format(stmt.capitalize(), step)
            line = next(
                line for line in result.splitlines() if line.endswith(comment)
            )
            assert line.split()[1] == str(nhits)


@pytest.mark.parametrize('n', [1, 2])
@pytest.mark.parametrize('legacy', [True, False])
def test_load_stats_files(legacy, n):
    """
    Test the loading of stats files. If ``legacy`` is true, the
    tempfiles are written from
    :py:class:`line_profiler._line_profiler.LineStats` objects instead
    of  :py:class:`line_profiler.line_profiler.LineStats` objects, so
    that we ensure that ``'.lprof'`` files written by old versions of
    :py:mod:`line_profiler` is still properly handled.
    """

    def write(stats, filename):
        if legacy:
            legacy_stats = type(stats).__base__(stats.timings, stats.unit)
            assert not isinstance(legacy_stats, LineStats)
            with open(filename, mode='wb') as fobj:
                pickle.dump(legacy_stats, fobj)
        else:
            stats.to_file(filename)
        return filename

    stats1 = LineStats({('foo', 1, 'spam.py'): [(2, 3, 3600)]}, 0.015625)
    stats2 = LineStats(
        {
            ('foo', 1, 'spam.py'): [(2, 4, 700)],
            ('bar', 10, 'spam.py'): [(10, 20, 1000)],
        },
        0.0625,
    )
    with TemporaryDirectory() as tmpdir:
        fname1 = write(stats1, os.path.join(tmpdir, '1.lprof'))
        if n == 1:
            stats_combined = stats1
            files = [fname1]
        else:
            fname2 = write(stats2, os.path.join(tmpdir, '2.lprof'))
            stats_combined = stats1 + stats2
            files = [fname1, fname2]
        stats_read = LineStats.from_files(*files)
    assert isinstance(stats_read, LineStats)
    assert stats_read == stats_combined



"""
Tests for :py:mod:`line_profiler.autoprofile.eager_preimports`.

Notes
-----
Most of the features are already covered by the doctests.
"""

from __future__ import annotations
import subprocess
import sys
from contextlib import ExitStack
from functools import partial
from pathlib import Path
from operator import methodcaller
from runpy import run_path
from tempfile import TemporaryDirectory
from textwrap import dedent
from types import SimpleNamespace
from typing import Collection, Generator, Sequence, Type, Optional, Union
from uuid import uuid4
from warnings import catch_warnings

import pytest

try:
    import flake8  # noqa: F401
except ImportError:
    HAS_FLAKE8 = False
else:
    HAS_FLAKE8 = True

from line_profiler.autoprofile.eager_preimports import (
    split_dotted_path,
    resolve_profiling_targets,
    write_eager_import_module,
)


def write(path: Path, content: Optional[str] = None) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    if content is None:
        path.touch()
    else:
        path.write_text(dedent(content).strip('\n'))


def gen_names(name) -> Generator[str, None, None]:
    while True:
        yield '_'.join([name, *str(uuid4()).split('-')])


@pytest.fixture
def preserve_sys_state() -> None:
    old_path = sys.path.copy()
    old_modules = sys.modules.copy()
    try:
        yield
    finally:
        sys.path.clear()
        sys.path[:] = old_path
        sys.modules.clear()
        sys.modules.update(old_modules)


@pytest.fixture
def sample_package(preserve_sys_state: None, tmp_path: Path) -> str:
    """
    Write a normal package and put it in :py:data:`sys.path`.  When
    we're done, reset :py:data:`sys.path` and `sys.modules`.
    """
    module_name = next(
        name for name in gen_names('my_sample_pkg') if name not in sys.modules
    )
    new_path = tmp_path / '_modules'
    write(new_path / module_name / '__init__.py')
    write(new_path / module_name / 'foo' / '__init__.py')
    write(new_path / module_name / 'foo' / 'bar.py')
    write(
        new_path / module_name / 'foo' / 'baz.py',
        """
          '''
          This is a bad module.
          '''
          raise AssertionError
          """,
    )
    write(new_path / module_name / 'foobar.py')
    # Cleanup managed with `preserve_sys_state()`
    sys.path.insert(0, str(new_path))
    yield module_name


@pytest.fixture
def sample_namespace_package(
    preserve_sys_state: None, tmp_path_factory: pytest.TempPathFactory
) -> str:
    """
    Write a namespace package and put it in :py:data:`sys.path`.  When
    we're done, reset :py:data:`sys.path` and `sys.modules`.
    """
    module_name = next(
        name
        for name in gen_names('my_sample_namespace_pkg')
        if name not in sys.modules
    )
    new_paths = [
        tmp_path_factory.mktemp('_modules-', numbered=True) for _ in range(3)
    ]
    for submod, new_path in zip(['one', 'two', 'three'], new_paths):
        write(new_path / module_name / (submod + '.py'))
        # Cleanup managed with `preserve_sys_state()`
        sys.path.insert(0, str(new_path))
    yield module_name


@pytest.mark.parametrize(
    ('adder', 'xc'),
    [('foo; bar', ValueError), (1, TypeError), ('(foo\n .bar)', ValueError)],
)
def test_write_eager_import_module_wrong_adder(
    adder: str, xc: Type[Exception]
) -> None:
    """
    Test passing an erroneous ``adder`` to
    :py:meth:`~.write_eager_import_module()`.
    """
    with pytest.raises(xc):
        write_eager_import_module(['foo'], adder=adder)


@pytest.mark.skipif(not HAS_FLAKE8, reason='no `flake8`')
def test_written_module_pep8_compliance(sample_package: str):
    """
    Test that the module written by
    :py:meth:`~.write_eager_import_module()` passes linting by
    :py:mod:`flake8`.
    """
    with TemporaryDirectory() as tmpdir:
        module = Path(tmpdir) / 'module.py'
        with module.open(mode='w') as fobj:
            write_eager_import_module(
                [sample_package + '.foobar'],
                recurse=[sample_package + '.foo'],
                stream=fobj,
            )
        print(module.read_text())
        (
            subprocess.run(
                [
                    sys.executable,
                    '-m',
                    'flake8',
                    '--extend-ignore=E501',  # Allow long lines
                    module,
                ]
            ).check_returncode()
        )


@pytest.mark.parametrize(
    ('dotted_paths', 'recurse', 'warnings', 'error'),
    [
        (
            ['__MODULE__.foobar'],
            ['__MODULE__.foo'],
            # `foo.baz` is indirectly included, so its raising an error
            # shouldn't cause the script to error out
            [{'target cannot', '__MODULE__.foo.baz'}],
            None,
        ),
        # We don't recurse down `__MODULE__.foo`, so that doesn't give a
        # warning; but `__MODULE__.baz` cannot be imported because it
        # doesn't exist
        (
            ['__MODULE__.foo', '__MODULE__.baz'],
            False,
            [{'target cannot', '__MODULE__.baz'}],
            None,
        ),
        # If we do recurse however, `__MODULE__.foo.baz` also ends up in
        # the warning
        # (also there's a `__MODULE___foo` which doesn't exist, about which
        # the warning is issued during module generation)
        (
            ['__MODULE__' + '_foo', '__MODULE__', '__MODULE__.baz'],
            True,
            [
                {'target cannot', '__MODULE__' + '_foo'},  # Fails at write
                {
                    'targets cannot',  # Fails at import
                    '__MODULE__.foo.baz',
                    '__MODULE__.baz',
                },
            ],
            None,
        ),
        # And if the problematic module is an explicit target, raise the
        # error
        (['__MODULE__', '__MODULE__.foo.baz'], False, [], AssertionError),
    ],
)
def test_written_module_error_handling(
    sample_package: str,
    dotted_paths: Collection[str],
    recurse: Union[Collection[str], bool],
    warnings: Sequence[Collection[str]],
    error: Union[Type[Exception], None],
):
    """
    Test that the module written by
    :py:meth:`~.write_eager_import_module()` gracefully handles errors
    for implicitly included modules.
    """
    replace = methodcaller('replace', '__MODULE__', sample_package)
    dotted_paths = [replace(target) for target in dotted_paths]
    if recurse not in (True, False):
        recurse = [replace(target) for target in recurse]
    warnings = [
        {replace(fragment) for fragment in fragments} for fragments in warnings
    ]
    with TemporaryDirectory() as tmpdir:
        module = Path(tmpdir) / 'module.py'
        with ExitStack() as stack:
            enter = stack.enter_context
            # Set up the warning capturing early so that we catch both
            # warnings at module-generation time and execution time
            captured_warnings = enter(catch_warnings(record=True))
            with module.open(mode='w') as fobj:
                write_eager_import_module(
                    dotted_paths, recurse=recurse, stream=fobj
                )
            print(module.read_text())
            if error is not None:
                enter(pytest.raises(error))
            # Just use a dummy object, no need to instantiate a profiler
            prof = SimpleNamespace(
                add_imported_function_or_module=lambda *_, **__: 0
            )
            run_path(str(module), {'profile': prof}, 'module')
    assert len(captured_warnings) == len(warnings)
    for warning, fragments in zip(captured_warnings, warnings):
        for fragment in fragments:
            assert fragment in str(warning.message)


def test_split_dotted_path_staticity() -> None:
    """
    Test `split_dotted_path()` with different values for `static`.
    """
    split = partial(split_dotted_path, 'os.path.abspath')
    # Static analysis has no idea of `os.path` members since `os.path`
    # is dynamically imported from e.g. `posixpath` or `ntpath`
    assert split(static=True) == ('os', 'path.abspath')
    # The import system knows of the already-imported module `os.path`
    assert split(static=False) == ('os.path', 'abspath')


def test_resolve_profiling_targets_staticity(
    sample_namespace_package: str,
) -> None:
    """
    Test subpackage/-module discovery with `resolve_profiling_targets()`
    with different values for `static`.
    """
    all_targets = {
        f'{sample_namespace_package}.{submod}'
        for submod in ['one', 'two', 'three']
    } | {sample_namespace_package}
    # Static analysis can't handle namespace packages
    resolve = partial(
        resolve_profiling_targets, [sample_namespace_package], recurse=True
    )
    static_result = resolve(static=True)
    assert set(static_result.targets) < all_targets, static_result
    # The import system successfully retrieves all submodules
    dyn_result = resolve(static=False)
    assert set(dyn_result.targets) == all_targets, dyn_result



"""
Statically port utilities from ubelt and xdocest need for the autoprofile
features.

Similar Scripts:
    ~/code/xdoctest/dev/maintain/port_ubelt_utils.py
    ~/code/mkinit/dev/maintain/port_ubelt_code.py
    ~/code/line_profiler/dev/maintain/port_utilities.py
"""

import ubelt as ub
import liberator
import re


def generate_util_static():
    lib = liberator.Liberator(verbose=0)

    import ubelt
    import xdoctest

    if 1:
        lib.add_dynamic(ubelt.util_import.modpath_to_modname)
        lib.add_dynamic(ubelt.util_import.modname_to_modpath)
        lib.expand(['ubelt'])

    if 1:
        lib.add_dynamic(xdoctest.static_analysis.package_modpaths)
        lib.expand(['xdoctest'])

        # # Hack because ubelt and xdoctest define this
        del lib.body_defs['xdoctest.utils.util_import._platform_pylib_exts']

    # lib.expand(['ubelt', 'xdoctest'])
    text = lib.current_sourcecode()

    """
    pip install rope
    pip install parso
    """

    prefix = ub.codeblock(
        '''
        """
        This file was autogenerated based on code in :py:mod:`ubelt` and
        :py:mod:`xdoctest` via dev/maintain/port_utilities.py in the
        line_profiler repo.
        """
        '''
    )

    # Remove doctest references to ubelt
    new_lines = []
    for line in text.split('\n'):
        if line.strip().startswith('>>> from ubelt'):
            continue
        if line.strip().startswith('>>> import ubelt as ub'):
            line = re.sub(
                '>>> .*', '>>> # xdoctest: +SKIP("ubelt dependency")', line
            )
        new_lines.append(line)

    text = '\n'.join(new_lines)
    text = prefix + '\n' + text + '\n'
    return text


def main():
    text = generate_util_static()
    print(ub.highlight_code(text, backend='rich'))

    import parso
    import line_profiler

    target_fpath = (
        ub.Path(line_profiler.__file__).parent
        / 'autoprofile'
        / 'util_static.py'
    )

    new_module = parso.parse(text)
    if target_fpath.exists():
        old_module = parso.parse(target_fpath.read_text())
        new_names = [
            child.name.value
            for child in new_module.children
            if child.type in {'funcdef', 'classdef'}
        ]
        old_names = [
            child.name.value
            for child in old_module.children
            if child.type in {'funcdef', 'classdef'}
        ]
        print(set(old_names) - set(new_names))
        print(set(new_names) - set(old_names))

    target_fpath.write_text(text)

    # Fixup formatting
    if 1:
        ub.cmd(['black', target_fpath])


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/line_profiler/dev/maintain/port_utilities.py
    """
    main()



from __future__ import annotations
import contextlib
import os
import re
import shlex
import subprocess
import sys
import tempfile
import unittest

import pytest
import ubelt as ub

from kernprof import main, ContextualProfile


def f(x):
    """A function."""
    y = x + 10
    return y


def g(x):
    """A generator."""
    y = yield x + 10
    yield y + 20


@pytest.mark.parametrize(
    'use_kernprof_exec, args, expected_output, expect_error',
    [
        ([False, ['-m'], '', True]),
        # `python -m kernprof`
        (False, ['-m', 'mymod'], '[__MYMOD__]', False),
        # `kernprof`
        (True, ['-m', 'mymod'], '[__MYMOD__]', False),
        (
            False,
            ['-m', 'mymod', '-p', 'bar'],
            "[__MYMOD__, '-p', 'bar']",
            False,
        ),
        # `-p bar` consumed by `kernprof`, `-p baz` are not
        (
            False,
            ['-p', 'bar', '-m', 'mymod', '-p', 'baz'],
            "[__MYMOD__, '-p', 'baz']",
            False,
        ),
        # Separator `--` broke off the remainder, so the requisite arg for
        # `-m` is not found and we error out
        (False, ['-p', 'bar', '-m', '--', 'mymod', '-p', 'baz'], '', True),
        # Separator `--` broke off the remainder, so `-m` is passed to the
        # script instead of being parsed as the module to execute
        (
            False,
            ['-p', 'bar', 'mymod.py', '--', '-m', 'mymod', '-p', 'baz'],
            "['mymod.py', '-m', 'mymod', '-p', 'baz']",
            False,
        ),
    ],
)
def test_kernprof_m_parsing(
    use_kernprof_exec, args, expected_output, expect_error
):
    """
    Test that `kernprof -m` behaves like `python -m` in that it requires
    an argument and cuts off everything after it, passing that along
    to the module to be executed.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        temp_dpath = ub.Path(tmpdir)
        mod = (temp_dpath / 'mymod.py').resolve()
        mod.write_text(
            ub.codeblock(
                """
            import sys


            if __name__ == '__main__':
                print(sys.argv)
            """
            )
        )
        if use_kernprof_exec:
            cmd = ['kernprof']
        else:
            cmd = [sys.executable, '-m', 'kernprof']
        proc = ub.cmd(cmd + args, cwd=temp_dpath, verbose=2)
    if expect_error:
        assert proc.returncode
        return
    else:
        proc.check_returncode()
    expected_output = re.escape(expected_output).replace(
        '__MYMOD__', "'.*{}'".format(re.escape(os.path.sep + mod.name))
    )
    assert re.match('^' + expected_output, proc.stdout)


@pytest.mark.skipif(
    sys.version_info[:2] < (3, 11),
    reason='no `@enum.global_enum` in Python '
    f'{".".join(str(v) for v in sys.version_info[:3])}',
)
@pytest.mark.parametrize(
    ('flags', 'profiled_main'),
    [
        ('-lv -p mymod', True),  # w/autoprofile
        ('-lv', True),  # w/o autoprofile
        ('-b', False),
    ],
)  # w/o line profiling
def test_kernprof_m_sys_modules(flags, profiled_main):
    """
    Test that `kernprof -m` is amenable to modules relying on the global
    `sys` state (e.g. those using `@enum.global_enum`).
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        temp_dpath = ub.Path(tmpdir)
        (temp_dpath / 'mymod.py').write_text(
            ub.codeblock(
                """
            import enum
            import os
            import sys


            @enum.global_enum
            class MyEnum(enum.Enum):
                FOO = 1
                BAR = 2


            @profile
            def main():
                x = FOO.value + BAR.value
                print(x)
                assert x == 3


            if __name__ == '__main__':
                main()
            """
            )
        )
        cmd = [
            sys.executable,
            '-m',
            'kernprof',
            *shlex.split(flags),
            '-m',
            'mymod',
        ]
        proc = ub.cmd(cmd, cwd=temp_dpath, verbose=2)
    proc.check_returncode()
    assert proc.stdout.startswith('3\n')
    assert ('Function: main' in proc.stdout) == profiled_main


@pytest.mark.parametrize('autoprof', [True, False])
@pytest.mark.parametrize('static', [True, False])
def test_kernprof_m_import_resolution(static, autoprof):
    """
    Test that `kernprof -m` resolves the module using static and dynamic
    as is appropriate (toggled by the undocumented environment variable
    :env:`LINE_PROFILER_STATIC_ANALYSIS`; note that static analysis
    doesn't handle namespace modules while dynamic resolution does.
    """
    code = ub.codeblock("""
    import enum
    import os
    import sys


    @profile
    def main():
        print('Hello world')


    if __name__ == '__main__':
        main()
    """)
    cmd = [sys.executable, '-m', 'kernprof', '-lv']
    if autoprof:
        # Remove the explicit decorator, and use the `--prof-mod` flag
        code = '\n'.join(
            line for line in code.splitlines() if '@profile' not in line
        )
        cmd += ['-p', 'my_namesapce_pkg.mysubmod']
    with tempfile.TemporaryDirectory() as tmpdir:
        temp_dpath = ub.Path(tmpdir)
        namespace_mod_path = temp_dpath / 'my_namesapce_pkg' / 'mysubmod.py'
        namespace_mod_path.parent.mkdir()
        namespace_mod_path.write_text(code)
        python_path = tmpdir
        if 'PYTHONPATH' in os.environ:
            python_path += ':' + os.environ['PYTHONPATH']
        env = {
            **os.environ,
            # Toggle use of static analysis
            'LINE_PROFILER_STATIC_ANALYSIS': str(bool(static)),
            # Add the tempdir to `sys.path`
            'PYTHONPATH': python_path,
        }
        cmd += ['-m', 'my_namesapce_pkg.mysubmod']
        proc = ub.cmd(cmd, cwd=temp_dpath, verbose=2, env=env)
    if static:
        assert proc.returncode
        assert proc.stderr.startswith('Could not find module')
    else:
        proc.check_returncode()
        assert proc.stdout.startswith('Hello world\n')
        assert 'Function: main' in proc.stdout


@pytest.mark.parametrize('error', [True, False])
@pytest.mark.parametrize(
    'args',
    ['', '-pmymod'],  # Normal execution / auto-profile
)
def test_kernprof_sys_restoration(capsys, error, args):
    """
    Test that `kernprof.main()` and
    `line_profiler.autoprofile.autoprofile.run()` (resp.) properly
    restores `sys.path` (resp. `sys.modules['__main__']`) on the way
    out.

    Notes
    -----
    The test is run in-process.
    """
    with contextlib.ExitStack() as stack:
        enter = stack.enter_context
        tmpdir = enter(tempfile.TemporaryDirectory())
        assert tmpdir not in sys.path
        temp_dpath = ub.Path(tmpdir)
        (temp_dpath / 'mymod.py').write_text(
            ub.codeblock(
                f"""
            import sys


            def main():
                # Mess up `sys.path`
                sys.path.append({tmpdir!r})
                # Output
                print(1)
                # Optionally raise an error
                if {error!r}:
                    raise Exception


            if __name__ == '__main__':
                main()
            """
            )
        )
        enter(ub.ChDir(tmpdir))
        if error:
            ctx = pytest.raises(BaseException)
        else:
            ctx = contextlib.nullcontext()
        old_modules = sys.modules.copy()
        try:
            old_main = sys.modules.get('__main__')
            with ctx:
                main(['-l', *shlex.split(args), '-m', 'mymod'])
            out, _ = capsys.readouterr()
            assert out.startswith('1')
            assert tmpdir not in sys.path
            assert sys.modules.get('__main__') is old_main
        finally:
            sys.modules.clear()
            sys.modules.update(old_modules)


@pytest.mark.parametrize(
    ('flags', 'expected_stdout', 'expected_stderr'),
    [
        (
            '',  # Neutral verbosity level
            {
                '^Output to stdout': True,
                r"^Wrote .* '.*script\.py\.lprof'": True,
                r'^Inspect results with:'
                '\n'
                r'\S*python\S* -m line_profiler .*script\.py\.lprof': True,
                r'line_profiler\.autoprofile\.autoprofile'
                r'\.run\(\n(?:.+,\n)*.*\)': False,
                r'^\[kernprof .*\]': False,
                r'^Function: main': False,
            },
            {'^Output to stderr': True},
        ),
        (
            '--view',  # Verbosity level 1 = `--view`
            {
                '^Output to stdout': True,
                r"^Wrote .* '.*script\.py\.lprof'": True,
                r'^Inspect results with:'
                '\n'
                r'\S*python\S* -m line_profiler .*script\.py\.lprof': False,
                r'line_profiler\.autoprofile\.autoprofile'
                r'\.run\(\n(?:.+,\n)*.*\)': False,
                r'^\[kernprof .*\]': False,
                r'^Function: main': True,
            },
            {'^Output to stderr': True},
        ),
        (
            '-vv',  # Verbosity level 2, show diagnostics
            {
                '^Output to stdout': True,
                r"^\[kernprof .*\] Wrote .* '.*script\.py\.lprof'": True,
                r'Inspect results with:'
                '\n'
                r'\S*python\S* -m line_profiler .*script\.py\.lprof': False,
                r'line_profiler\.autoprofile\.autoprofile'
                r'\.run\(\n(?:.+,\n)*.*\)': True,
                r'^Function: main': True,
            },
            {'^Output to stderr': True},
        ),
        # Verbosity level -1, suppress `kernprof` output
        (
            '--quiet',
            {'^Output to stdout': True, 'Wrote': False},
            {'^Output to stderr': True},
        ),
        # Verbosity level -2, suppress script stdout
        # (also test verbosity arithmatics)
        ('--quiet --quiet --verbose -q', None, {'^Output to stderr': True}),
        # Verbosity level -3, suppress script stderr
        (
            '-qq --quiet',
            None,
            # This should have been `None`, but there's something weird with
            # `coverage` in CI which causes a spurious warning...
            {'^Output to stderr': False},
        ),
    ],
)
def test_kernprof_verbosity(flags, expected_stdout, expected_stderr):
    """
    Test the various verbosity levels of `kernprof`.
    """
    with contextlib.ExitStack() as stack:
        enter = stack.enter_context
        tmpdir = enter(tempfile.TemporaryDirectory())
        temp_dpath = ub.Path(tmpdir)
        (temp_dpath / 'script.py').write_text(
            ub.codeblock(
                """
            import sys


            def main():
                print('Output to stdout', file=sys.stdout)
                print('Output to stderr', file=sys.stderr)


            if __name__ == '__main__':
                main()
            """
            )
        )
        enter(ub.ChDir(tmpdir))
        proc = ub.cmd(
            [
                'kernprof',
                '-l',
                # Add an eager pre-import target
                '-p',
                'script.py',
                '-p',
                'zipfile',
                '-z',
                *shlex.split(flags),
                'script.py',
            ]
        )
    proc.check_returncode()
    print(proc.stdout)
    for expected_outputs, stream in [
        (expected_stdout, proc.stdout),
        (expected_stderr, proc.stderr),
    ]:
        if expected_outputs is None:
            assert not stream
            continue
        for pattern, expect_match in expected_outputs.items():
            found = re.search(pattern, stream, flags=re.MULTILINE)
            if not bool(found) == expect_match:
                msg = ub.paragraph(
                    f"""
                    Searching for pattern: {pattern!r} in output.
                    Did we expect a match? {expect_match!r}.
                    Did we get a match? {bool(found)!r}.
                    """
                )
                raise AssertionError(msg)


def test_kernprof_eager_preimport_bad_module():
    """
    Test for the preservation of the full traceback when an error occurs
    in an auto-generated pre-import module.
    """
    bad_module = """raise Exception('Boo')"""
    with contextlib.ExitStack() as stack:
        enter = stack.enter_context
        tmpdir = enter(tempfile.TemporaryDirectory())
        temp_dpath = ub.Path(tmpdir)
        (temp_dpath / 'my_bad_module.py').write_text(bad_module)
        enter(ub.ChDir(tmpdir))
        python_path = os.environ.get('PYTHONPATH', '')
        if python_path:
            python_path = f'{python_path}:{tmpdir}'
        else:
            python_path = tmpdir
        proc = ub.cmd(
            [
                'kernprof',
                '-l',
                # Add an eager pre-import target
                '-pmy_bad_module',
                '-c',
                'print(1)',
            ],
            env={**os.environ, 'PYTHONPATH': python_path},
        )
    # Check that the traceback is preserved
    print(proc.stdout)
    print(proc.stderr, file=sys.stderr)
    assert proc.returncode
    assert 'import my_bad_module' in proc.stderr
    assert bad_module in proc.stderr
    # Check that the generated tempfiles are wiped
    reverse_iter_lines = iter(reversed(proc.stderr.splitlines()))
    next(line for line in reverse_iter_lines if 'import my_bad_module' in line)
    tb_header = next(reverse_iter_lines).strip()
    match = re.match('File ([\'"])(.+)\\1, line [0-9]+, in .*', tb_header)
    assert match
    tmp_mod = match.group(2)
    assert not os.path.exists(tmp_mod)
    assert not os.path.exists(os.path.dirname(tmp_mod))


@pytest.mark.parametrize('stdin', [True, False])
def test_kernprof_bad_temp_script(stdin):
    """
    Test for the preservation of the full traceback when an error occurs
    in a temporary script supplied via `kernprof -c` or `kernprof -`.
    """
    bad_script = """1 / 0"""
    with contextlib.ExitStack() as stack:
        enter = stack.enter_context
        enter(ub.ChDir(enter(tempfile.TemporaryDirectory())))
        if stdin:
            proc = subprocess.run(
                ['kernprof', '-'],
                input=bad_script,
                capture_output=True,
                text=True,
            )
        else:
            proc = subprocess.run(
                ['kernprof', '-c', bad_script], capture_output=True, text=True
            )
    # Check that the traceback is preserved
    print(proc.stdout)
    print(proc.stderr, file=sys.stderr)
    assert proc.returncode
    assert '1 / 0' in proc.stderr
    assert 'ZeroDivisionError' in proc.stderr
    # Check that the generated tempfiles are wiped
    reverse_iter_lines = iter(reversed(proc.stderr.splitlines()))
    next(line for line in reverse_iter_lines if '1 / 0' in line)
    tb_header = next(reverse_iter_lines).strip()
    match = re.match('File ([\'"])(.+)\\1, line [0-9]+, in .*', tb_header)
    assert match
    tmp_script = match.group(2)
    assert not os.path.exists(tmp_script)
    assert not os.path.exists(os.path.dirname(tmp_script))


@pytest.mark.parametrize('debug', [True, False])
def test_bad_prof_mod_target(debug):
    """
    Test the handling of bad paths in `--prof-mod` targets.
    """
    with contextlib.ExitStack() as stack:
        enter = stack.enter_context
        enter(ub.ChDir(enter(tempfile.TemporaryDirectory())))
        proc = ub.cmd(
            [
                'kernprof',
                '-l',
                '-p',
                './nonexistent.py',
                '-c',
                'print("Output: foo")',
            ],
            env={**os.environ, 'LINE_PROFILER_DEBUG': str(debug)},
        )
        print(proc.stdout)
        print(proc.stderr, file=sys.stderr)
        proc.check_returncode()
        assert os.listdir()  # Profile results
    assert 'Output: foo' in proc.stdout
    assert re.search(r"1 .* target .*: \['\./nonexistent\.py'\]", proc.stderr)


@pytest.mark.parametrize('builtin', [True, False])
@pytest.mark.parametrize('module', [True, False])
def test_call_with_diagnostics(module, builtin):
    """
    Test the output of call signatures in debug messages.
    """
    to_run = ['-m', 'calendar'] if module else ['-c', 'print("Output: foo")']
    with contextlib.ExitStack() as stack:
        enter = stack.enter_context
        enter(ub.ChDir(enter(tempfile.TemporaryDirectory())))
        cmd = ['kernprof']
        if builtin:
            cmd += ['-b']
        proc = ub.cmd(
            cmd + to_run, env={**os.environ, 'LINE_PROFILER_DEBUG': 'true'}
        )
        print(proc.stdout)
        print(proc.stderr, file=sys.stderr)
        proc.check_returncode()
        assert os.listdir()  # Profile results
    has_runctx_call = re.search(
        r'Calling: .+\.runctx\(.+\)', proc.stdout, flags=re.DOTALL
    )
    has_execfile_call = re.search(
        r'execfile\(.+\)', proc.stdout, flags=re.DOTALL
    )
    assert bool(has_runctx_call) == (not builtin)
    assert bool(has_execfile_call) == (not module)


class TestKernprof(unittest.TestCase):
    def test_enable_disable(self):
        profile = ContextualProfile()
        self.assertEqual(profile.enable_count, 0)
        profile.enable_by_count()
        self.assertEqual(profile.enable_count, 1)
        profile.enable_by_count()
        self.assertEqual(profile.enable_count, 2)
        profile.disable_by_count()
        self.assertEqual(profile.enable_count, 1)
        profile.disable_by_count()
        self.assertEqual(profile.enable_count, 0)
        profile.disable_by_count()
        self.assertEqual(profile.enable_count, 0)

        with profile:
            self.assertEqual(profile.enable_count, 1)
            with profile:
                self.assertEqual(profile.enable_count, 2)
            self.assertEqual(profile.enable_count, 1)
        self.assertEqual(profile.enable_count, 0)

        with self.assertRaises(RuntimeError):
            self.assertEqual(profile.enable_count, 0)
            with profile:
                self.assertEqual(profile.enable_count, 1)
                raise RuntimeError()
        self.assertEqual(profile.enable_count, 0)

    def test_function_decorator(self):
        profile = ContextualProfile()
        f_wrapped = profile(f)
        self.assertEqual(f_wrapped.__name__, f.__name__)
        self.assertEqual(f_wrapped.__doc__, f.__doc__)

        self.assertEqual(profile.enable_count, 0)
        value = f_wrapped(10)
        self.assertEqual(profile.enable_count, 0)
        self.assertEqual(value, f(10))

    def test_gen_decorator(self):
        profile = ContextualProfile()
        g_wrapped = profile(g)
        self.assertEqual(g_wrapped.__name__, g.__name__)
        self.assertEqual(g_wrapped.__doc__, g.__doc__)

        self.assertEqual(profile.enable_count, 0)
        i = g_wrapped(10)
        self.assertEqual(profile.enable_count, 0)
        self.assertEqual(next(i), 20)
        self.assertEqual(profile.enable_count, 0)
        self.assertEqual(i.send(30), 50)
        self.assertEqual(profile.enable_count, 0)

        with self.assertRaises((StopIteration, RuntimeError)):
            next(i)
        self.assertEqual(profile.enable_count, 0)


if __name__ == '__main__':
    """
    CommandLine:
        python ./tests/test_kernprof.py
    """
    unittest.main()



"""

AutoProfile Script Demo
=======================

The following demo is end-to-end bash code that writes a demo script and
profiles it with autoprofile.

.. code:: bash

    # Write demo python script to disk
    python -c "if 1:
        import textwrap
        text = textwrap.dedent(
            '''
            def plus(a, b):
                return a + b

            def fib(n):
                a, b = 0, 1
                while a < n:
                    a, b = b, plus(a, b)

            def main():
                import math
                import time
                start = time.time()

                print('start calculating')
                while time.time() - start < 1:
                    fib(10)
                    math.factorial(1000)
                print('done calculating')

            main()
            '''
        ).strip()
        with open('demo.py', 'w') as file:
            file.write(text)
    "

    echo "---"
    echo "## Profile With AutoProfile"
    python -m kernprof -p demo.py -l demo.py
    python -m line_profiler -rmt demo.py.lprof
"""

from __future__ import annotations
import contextlib
import functools
import importlib.util
import operator
import sys
import types
from collections.abc import MutableMapping
from typing import Any, cast, Dict, Mapping
from typing import ContextManager
from .ast_tree_profiler import AstTreeProfiler
from .run_module import AstTreeModuleProfiler
from .line_profiler_utils import add_imported_function_or_module
from .util_static import modpath_to_modname

PROFILER_LOCALS_NAME = 'prof'


def _extend_line_profiler_for_profiling_imports(prof: Any) -> None:
    """Allow profiler to handle functions/methods, classes & modules with a single call.

    Add a method to LineProfiler that can identify whether the object is a
    function/method, class or module and handle it's profiling accordingly.
    Mainly used for profiling objects that are imported.
    (Workaround to keep changes needed by autoprofile separate from base LineProfiler)

    Args:
        prof (LineProfiler):
            instance of LineProfiler.
    """
    prof.add_imported_function_or_module = types.MethodType(
        add_imported_function_or_module, prof
    )


def run(
    script_file: str,
    ns: MutableMapping[str, Any],
    prof_mod: list[str],
    profile_imports: bool = False,
    as_module: bool = False,
) -> None:
    """Automatically profile a script and run it.

    Profile functions, classes & modules specified in prof_mod without needing to add
    @profile decorators.

    Args:
        script_file (str):
            path to script being profiled.

        ns (dict):
            "locals" from kernprof scope.

        prof_mod (List[str]):
            list of imports to profile in script.
            passing the path to script will profile the whole script.
            the objects can be specified using its dotted path or full path (if applicable).

        profile_imports (bool):
            if True, when auto-profiling whole script, profile all imports aswell.

        as_module (bool):
            Whether we're running script_file as a module
    """

    class restore_dict:
        def __init__(self, d: MutableMapping[str, Any], target=None):
            self.d = d
            self.target = target
            self.copy: Mapping[str, Any] | None = None

        def __enter__(self):
            assert self.copy is None
            self.copy = dict(self.d)
            return self.target

        def __exit__(self, *_, **__):
            self.d.clear()
            if self.copy is not None:
                self.d.update(self.copy)
            self.copy = None

    Profiler: type[AstTreeModuleProfiler] | type[AstTreeProfiler]
    namespace: MutableMapping[str, Any]
    ctx: ContextManager

    if as_module:
        Profiler = AstTreeModuleProfiler
        module_name = modpath_to_modname(script_file)
        if not module_name:
            raise ModuleNotFoundError(
                f'script_file = {script_file!r}: cannot find corresponding module'
            )

        module_obj = types.ModuleType(module_name)
        namespace = vars(module_obj)
        namespace.update(ns)

        # Set the `__spec__` correctly
        module_obj.__spec__ = importlib.util.find_spec(module_name)

        # Set the module object to `sys.modules` via a callback, and
        # then restore it via the context manager
        callback = functools.partial(
            operator.setitem, sys.modules, '__main__', module_obj
        )
        ctx = restore_dict(sys.modules, callback)
    else:
        Profiler = AstTreeProfiler
        namespace = ns
        ctx = contextlib.nullcontext(lambda: None)

    profiler = Profiler(script_file, prof_mod, profile_imports)
    tree_profiled = profiler.profile()

    _extend_line_profiler_for_profiling_imports(ns[PROFILER_LOCALS_NAME])
    code_obj = compile(tree_profiled, script_file, 'exec')
    with ctx as callback:
        callback()
        exec(code_obj, cast(Dict[str, Any], namespace), namespace)



# cython: profile=True
# cython: linetrace=True
# distutils: define_macros=CYTHON_TRACE=1


ctypedef long double Float


def cos(Float x, int n):  # Start: cos
    cdef Float neg_xsq = -x * x
    cdef Float last_term = 1.
    cdef Float result = 0.
    for n in range(2, 2 * n, 2):
        result += last_term
        last_term *= neg_xsq / <Float>(n * (n - 1))
    return result + last_term   # End: cos


cpdef sin(Float x, int n):  # Start: sin
    cdef Float neg_xsq = -x * x
    cdef Float last_term = x
    cdef Float result = 0.
    for n in range(2, 2 * n, 2):
        result += last_term
        last_term *= neg_xsq / <Float>(n * (n + 1))
    return result + last_term  # End: sin



from __future__ import annotations

from typing import Any, Mapping

class LineStats:
    timings: Mapping[tuple[str, int, str], list[tuple[int, int, int]]]
    unit: float

    def __init__(
        self,
        timings: Mapping[tuple[str, int, str], list[tuple[int, int, int]]],
        unit: float,
    ) -> None: ...

class LineProfiler:
    def enable_by_count(self) -> None: ...
    def disable_by_count(self) -> None: ...
    def add_function(self, func: Any) -> None: ...
    def get_stats(self) -> LineStats: ...
    def dump_stats(self, filename: str) -> None: ...

def label(code: Any) -> Any: ...



#!/usr/bin/env python
from os.path import exists
import sys
import os
import warnings
import setuptools


def _choose_build_method():
    DISABLE_C_EXTENSIONS = os.environ.get('DISABLE_C_EXTENSIONS', '').lower()
    LINE_PROFILER_BUILD_METHOD = os.environ.get(
        'LINE_PROFILER_BUILD_METHOD', 'auto'
    ).lower()

    if DISABLE_C_EXTENSIONS in {'true', 'on', 'yes', '1'}:
        LINE_PROFILER_BUILD_METHOD = 'setuptools'

    if LINE_PROFILER_BUILD_METHOD == 'auto':
        try:
            import Cython  # NOQA
        except ImportError:
            try:
                import skbuild  # NOQA
                import cmake  # NOQA
                import ninja  # NOQA
            except ImportError:
                # The main fallback disables c-extensions
                LINE_PROFILER_BUILD_METHOD = 'setuptools'
            else:
                # This should never be hit
                LINE_PROFILER_BUILD_METHOD = 'scikit-build'
        else:
            # Use plain cython by default
            LINE_PROFILER_BUILD_METHOD = 'cython'

    return LINE_PROFILER_BUILD_METHOD


def parse_version(fpath):
    """
    Statically parse the version number from a python file
    """
    value = static_parse('__version__', fpath)
    return value


def static_parse(varname, fpath):
    """
    Statically parse the a constant variable from a python file.
    Raise an error if the variable is not a constant.
    """
    import ast

    if not exists(fpath):
        raise ValueError('fpath={!r} does not exist'.format(fpath))
    with open(fpath, 'r') as file_:
        sourcecode = file_.read()
    pt = ast.parse(sourcecode)

    class StaticVisitor(ast.NodeVisitor):
        def visit_Assign(self, node: ast.Assign):
            for target in node.targets:
                if getattr(target, 'id', None) == varname:
                    value: ast.expr = node.value
                    if not isinstance(value, ast.Constant):
                        raise ValueError(
                            'variable {!r} is not a constant'.format(varname)
                        )
                    self.static_value = value.value

    visitor = StaticVisitor()
    visitor.visit(pt)
    try:
        value = visitor.static_value
    except AttributeError:
        value = 'Unknown {}'.format(varname)
        warnings.warn(value)
    return value


def parse_description():
    """
    Parse the description in the README file

    CommandLine:
        pandoc --from=markdown --to=rst --output=README.rst README.md
        python -c "import setup; print(setup.parse_description())"
    """
    from os.path import dirname, join, exists

    readme_fpath = join(dirname(__file__), 'README.rst')
    # This breaks on pip install, so check that it exists.
    if exists(readme_fpath):
        with open(readme_fpath, 'r') as f:
            text = f.read()
        return text
    return ''


def parse_requirements(fname='requirements.txt', versions=False):
    """
    Parse the package dependencies listed in a requirements file but strips
    specific versioning information.

    Args:
        fname (str): path to requirements file
        versions (bool | str, default=False):
            If true include version specs.
            If strict, then pin to the minimum version.

    Returns:
        List[str]: list of requirements items
    """
    from os.path import exists, dirname, join
    import re

    require_fpath = fname

    def parse_line(line, dpath=''):
        """
        Parse information from a line in a requirements text file

        line = 'git+https://a.com/somedep@sometag#egg=SomeDep'
        line = '-e git+https://a.com/somedep@sometag#egg=SomeDep'
        """
        # Remove inline comments
        comment_pos = line.find(' #')
        if comment_pos > -1:
            line = line[:comment_pos]

        if line.startswith('-r '):
            # Allow specifying requirements in other files
            target = join(dpath, line.split(' ')[1])
            for info in parse_require_file(target):
                yield info
        else:
            # See: https://www.python.org/dev/peps/pep-0508/
            info = {'line': line}
            if line.startswith('-e '):
                info['package'] = line.split('#egg=')[1]
            else:
                if ';' in line:
                    pkgpart, platpart = line.split(';')
                    # Handle platform specific dependencies
                    # setuptools.readthedocs.io/en/latest/setuptools.html
                    # #declaring-platform-specific-dependencies
                    plat_deps = platpart.strip()
                    info['platform_deps'] = plat_deps
                else:
                    pkgpart = line
                    platpart = None

                # Remove versioning from the package
                pat = '(' + '|'.join(['>=', '==', '>']) + ')'
                parts = re.split(pat, pkgpart, maxsplit=1)
                parts = [p.strip() for p in parts]

                info['package'] = parts[0]
                if len(parts) > 1:
                    op, rest = parts[1:]
                    version = rest  # NOQA
                    info['version'] = (op, version)
            yield info

    def parse_require_file(fpath):
        dpath = dirname(fpath)
        with open(fpath, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if line and not line.startswith('#'):
                    for info in parse_line(line, dpath=dpath):
                        yield info

    def gen_packages_items():
        if exists(require_fpath):
            for info in parse_require_file(require_fpath):
                parts = [info['package']]
                if versions and 'version' in info:
                    if versions == 'strict':
                        # In strict mode, we pin to the minimum version
                        if info['version']:
                            # Only replace the first >= instance
                            verstr = ''.join(info['version']).replace(
                                '>=', '==', 1
                            )
                            parts.append(verstr)
                    else:
                        parts.extend(info['version'])
                if not sys.version.startswith('3.4'):
                    # apparently package_deps are broken in 3.4
                    plat_deps = info.get('platform_deps')
                    if plat_deps is not None:
                        parts.append(';' + plat_deps)
                item = ''.join(parts)
                yield item

    packages = list(gen_packages_items())
    return packages


long_description = """\
line_profiler will profile the time individual lines of code take to execute.
The profiler is implemented in C via Cython in order to reduce the overhead of
profiling.

Also included is the script kernprof.py which can be used to conveniently
profile Python applications and scripts either with line_profiler or with the
function-level profiling tools in the Python standard library.
"""


NAME = 'line_profiler'
INIT_PATH = 'line_profiler/line_profiler.py'
VERSION = parse_version(INIT_PATH)


if __name__ == '__main__':
    setupkw = {}

    LINE_PROFILER_BUILD_METHOD = _choose_build_method()
    if LINE_PROFILER_BUILD_METHOD == 'setuptools':
        setup = setuptools.setup
    elif LINE_PROFILER_BUILD_METHOD == 'scikit-build':
        import skbuild  # NOQA

        setup = skbuild.setup
    elif LINE_PROFILER_BUILD_METHOD == 'cython':
        # no need to try importing cython because an import
        # was already attempted in _choose_build_method
        import multiprocessing
        from setuptools import Extension
        from Cython.Build import cythonize

        def run_cythonize(force=False):
            return cythonize(
                Extension(
                    name='line_profiler._line_profiler',
                    sources=[
                        'line_profiler/_line_profiler.pyx',
                        'line_profiler/timers.c',
                        'line_profiler/c_trace_callbacks.c',
                    ],
                    # force a recompile if this changes
                    depends=[
                        'line_profiler/_map_helpers.pxd',
                    ],
                    language='c++',
                    define_macros=[
                        (
                            'CYTHON_TRACE',
                            (1 if os.getenv('DEV') == 'true' else 0),
                        )
                    ],
                ),
                compiler_directives={
                    'language_level': 3,
                    'infer_types': True,
                    'legacy_implicit_noexcept': True,
                    'linetrace': (
                        True if os.getenv('DEV') == 'true' else False
                    ),
                },
                include_path=[
                    'line_profiler/python25.pxd',
                    'line_profiler/_map_helpers.pxd',
                ],
                force=force,
                nthreads=multiprocessing.cpu_count(),
            )

        setupkw.update(dict(ext_modules=run_cythonize()))
        setup = setuptools.setup
    else:
        raise Exception('Unknown build method')

    setupkw['install_requires'] = parse_requirements(
        'requirements/runtime.txt', versions='loose'
    )
    setupkw['extras_require'] = {
        'all': parse_requirements('requirements.txt', versions='loose'),
        'tests': parse_requirements('requirements/tests.txt', versions='loose'),
        'optional': parse_requirements(
            'requirements/optional.txt', versions='loose'
        ),
        'all-strict': parse_requirements('requirements.txt', versions='strict'),
        'runtime-strict': parse_requirements(
            'requirements/runtime.txt', versions='strict'
        ),
        'tests-strict': parse_requirements(
            'requirements/tests.txt', versions='strict'
        ),
        'optional-strict': parse_requirements(
            'requirements/optional.txt', versions='strict'
        ),
        'ipython': parse_requirements(
            'requirements/ipython.txt', versions='loose'
        ),
        'ipython-strict': parse_requirements(
            'requirements/ipython.txt', versions='strict'
        ),
    }
    setupkw['entry_points'] = {
        'console_scripts': [
            'kernprof=kernprof:main',
        ],
    }
    setupkw['name'] = NAME
    setupkw['version'] = VERSION
    setupkw['author'] = 'Robert Kern'
    setupkw['author_email'] = 'robert.kern@enthought.com'
    setupkw['url'] = 'https://github.com/pyutils/line_profiler'
    setupkw['description'] = 'Line-by-line profiler'
    setupkw['long_description'] = parse_description()
    setupkw['long_description_content_type'] = 'text/x-rst'
    setupkw['license'] = 'BSD'
    setupkw['packages'] = list(setuptools.find_packages())
    setupkw['py_modules'] = ['kernprof', 'line_profiler']
    setupkw['python_requires'] = '>=3.8'
    setupkw['license_files'] = ['LICENSE.txt', 'LICENSE_Python.txt']
    setupkw['package_data'] = {'line_profiler': ['py.typed', '*.pyi', '*.toml']}
    # `include_package_data` is needed to put `rc/line_profiler.toml` in
    # the wheel
    setupkw['include_package_data'] = True
    setupkw['keywords'] = [
        'timing',
        'timer',
        'profiling',
        'profiler',
        'line_profiler',
    ]
    setupkw['classifiers'] = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: C',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development',
    ]
    setup(**setupkw)



from __future__ import annotations


def test_import():
    import line_profiler

    assert hasattr(line_profiler, 'LineProfiler')
    assert hasattr(line_profiler, '__version__')


def test_version():
    import line_profiler
    from packaging.version import Version
    import kernprof

    line_profiler_version1 = Version(line_profiler.__version__)
    line_profiler_version2 = Version(line_profiler.line_profiler.__version__)
    kernprof_version = Version(kernprof.__version__)
    assert (
        line_profiler_version1 == line_profiler_version2 == kernprof_version
    ), 'All 3 places should have the same version'
